@@ -1,5 +1,7 @@
 import msgpack
+import traceback
 from collections import deque
+from threading import Condition, Lock
 from time import time
 from mixins import mixins
 from util import VimError
@@ -20,6 +22,39 @@ class Remote(object):
         return hasattr(other, '_handle') and self._handle == other._handle
 
 
+class BorrowedStreamMutex(object):
+    """
+    Helper class used to safely acquire or "borrow" the stream mutex from a
+    thread that is blocking on a `next_event` call
+    """
+    def __init__(self, stream, stream_mutex, switcher):
+        self.stream = stream
+        self.stream_mutex = stream_mutex
+        self.switcher = switcher
+        self.inner_mutex = Lock()
+
+    def __enter__(self):
+        self.inner_mutex.acquire()
+        if not self.stream_mutex.acquire(False):
+            # Another thread is currently blocking on a 'next_event' call,
+            # we need to interrupt it and try to acquire `stream_mutex`
+            # ourselves, or else we would have to wait until an event is
+            # received
+            with self.switcher:
+                # Send the interrupt
+                self.stream.interrupt()
+                # Assert that the other thread was interrupted in time before
+                # waiting on the switcher condition
+                if not self.stream_mutex.acquire(False):
+                    self.switcher.wait()
+                    self.stream_mutex.acquire()
+                    self.switcher.notify()
+
+    def __exit__(self, type, value, traceback):
+        self.stream_mutex.release()
+        self.inner_mutex.release()
+
+
 class Client(object):
     """
     Neovim client. It depends on a stream, an object that implements two
@@ -29,17 +64,32 @@ class Client(object):
 
     Both methods should be fully blocking.
     """
+
+
     def __init__(self, stream):
         self._request_id = 0
-        self._stream = stream
-        self._unpacker = msgpack.Unpacker()
+        self.stream = stream
+        self.unpacker = msgpack.Unpacker()
         self.pending_events = deque()
         self.vim = None
+        # These are the constructs used to synchronize access:
+        #
+        # - request_mutex: mutual exclusion on the `msgpack_rpc_request` method
+        # - stream_mutex: mutual exlusion on the stream. It must be acquired
+        #   by both `msgpack_rpc_request` and `next_event`.
+        # - switcher: Helper mutex/condition used to pass `stream_mutex` from a
+        # thread blocking on `next_event` to another thread calling
+        # `msgpack_rpc_request`
+        self.stream_mutex = Lock()
+        self.switcher = Condition(Lock())
+        self.borrowed_stream_mutex = BorrowedStreamMutex(self.stream,
+                                                     self.stream_mutex,
+                                                     self.switcher)
 
     def pop_events(self, max_index=None):
         if max_index is None:
             rv = list(self.pending_events)
-            self.pending_events = deque()
+            self.pending_events.clear()
         else:
             rv = []
             i = 0
@@ -55,47 +105,71 @@ class Client(object):
         """
         while True:
             try:
-                return self._unpacker.next()
+                return self.unpacker.next()
             except StopIteration:
-                chunk = self._stream.read(timeout)
+                chunk = self.stream.read(timeout)
                 if not chunk:
                     return
-                self._unpacker.feed(chunk)
+                self.unpacker.feed(chunk)
 
     def msgpack_rpc_request(self, method_id, params):
         """
         Sends a msgpack-rpc request to Neovim and returns the response
         """
-        # increment request id
-        request_id = self._request_id + 1
-        self._stream.write(msgpack.packb([0, request_id, method_id, params]))
-        # Enter a loop feeding the unpacker with data until we parse
-        # the response
-        message = None
-        while not message:
-            message = self.next_message()
-            if message[0] == 2:
-                # event, add to the pending queue
-                self.pending_events.append(message[1:])
-                message = None
-        # update request id
-        self._request_id = request_id
-        return message
+        with self.borrowed_stream_mutex:
+            request_id = self._request_id + 1
+            # Send the request
+            self.stream.write(
+                msgpack.packb([0, request_id, method_id, params]))
+            # Enter a loop feeding the unpacker with data until we parse the
+            # response
+            message = None
+            while not message:
+                message = self.next_message()
+                if message and message[0] == 2:
+                    # event, add to the pending queue
+                    self.pending_events.append(message[1:])
+                    message = None
+            # Update request id
+            self._request_id = request_id
+            return message
 
     def next_event(self, timeout=None):
         """
         Returns the next server event
         """
-        if len(self.pending_events):
-            return self.pending_events.popleft()
-        message = self.next_message(timeout)
-        # The message type must be 2, which is the msgpack-rpc notification
-        # type(http://wiki.msgpack.org/display/MSGPACK/RPC+specification). The
-        # only other possible message type the server will send is a
-        # response(1), but those must be received from `msgpack_rpc_request`
-        assert message[0] == 2
-        if message:
-            return message[1:]
+        with self.stream_mutex:
+            while True:
+                if len(self.pending_events):
+                    return self.pending_events.popleft()
+                message = self.next_message(timeout)
+                with self.switcher:
+                    if not message:
+                        # If interrupted by a `msgpack_rpc_request` call from
+                        # another thread, wait for a signal and try again
+                        self.stream_mutex.release()
+                        self.switcher.notify()
+                        self.switcher.wait()
+                        self.stream_mutex.acquire()
+                        continue
+                    # The message type must be 2, which is the msgpack-rpc
+                    # notification type
+                    # (http://wiki.msgpack.org/display/MSGPACK/RPC+specification).
+                    # The only other possible message type the server will send
+                    # is a response(1), but those must be received from
+                    # `msgpack_rpc_request`
+                    assert not message or message[0] == 2
+                    if message:
+                        return message[1:]
+
+    def push_event(self, event_name, event_data):
+        """
+        Pushes a "virtual event" that will be returned by `next_event`.  This
+        method can called from other threads for integration with event loops,
+        such as those provided by GUI libraries.
+        """
+        with self.borrowed_stream_mutex:
+            self.pending_events.append([event_name, event_data])
 
     def expect(self, event_type, timeout=None, predicate=lambda e: True):
         """
@@ -119,6 +193,9 @@ class Client(object):
             # block until the first line redraws with the string 'test'
             vim.expect('redraw', timeout=2, lambda e: e['lines'][0] == 'test')
         ```
+
+        Warning: This function probably should not be called in multithreaded
+        programs
         """
         for i, event in enumerate(self.pending_events):
             if event[0] == event_type and predicate(event):
@@ -135,8 +212,6 @@ class Client(object):
                 return self.pop_events()
             timeout -= time() - start
 
-        raise Exception('Timed out')
-
     def discover_api(self):
         """
         Discovers the remote API using the special method '0'. After this
@@ -153,6 +228,8 @@ class Client(object):
         classes = {'vim': type('Vim', (), {})}
         setattr(classes['vim'], 'next_event',
                 lambda s, *args, **kwargs: self.next_event(*args, **kwargs))
+        setattr(classes['vim'], 'push_event',
+                lambda s, *args, **kwargs: self.push_event(*args, **kwargs))
         setattr(classes['vim'], 'expect',
                 lambda s, *args, **kwargs: self.expect(*args, **kwargs))
         # Build classes for manipulating the remote structures, assigning to a
@@ -179,7 +256,8 @@ class Client(object):
             classes[name] = type(mixin.__name__, (classes[name], mixin,), {})
         # Create the 'vim object', which is a singleton of the 'Vim' class
         self.vim = classes['vim']()
-        self.vim.channel_id = channel_id
+        # Initialize with some useful attributes
+        classes['vim'].initialize(self.vim, classes, channel_id)
         # Add attributes for each other class
         for name, klass in classes.items():
             if name != 'vim':
@@ -216,7 +294,11 @@ def generate_wrapper(client, klass, name, fid, return_type, parameters):
             raise VimError(result[2])
         if hasattr(client.vim, return_type):
             # result should be a handle, wrap in it's specialized class
-            return getattr(client.vim, return_type)(client.vim, result[3])
+            klass = getattr(client.vim, return_type)
+            rv = klass(client.vim, result[3])
+            klass.initialize(rv)
+            return rv
+
         return result[3]
     setattr(klass, name, rv)
 
@@ -229,3 +311,5 @@ def fname(name):
         f.__name__ = name
         return f
     return dec
+
+
