@@ -1,3 +1,4 @@
+from collections import deque
 import sys, pyuv
 
 class UvStream(object):
@@ -8,7 +9,7 @@ class UvStream(object):
     def __init__(self, address=None, port=None):
         self._loop = pyuv.Loop()
         self._error = None
-        self._data = None
+        self._data = deque()
         self._written = True
         self._connected = False
         self._timed_out = False
@@ -32,6 +33,9 @@ class UvStream(object):
             self._read_stream.open(sys.stdin.fileno())
             self._write_stream = pyuv.Pipe(self._loop) 
             self._write_stream.open(sys.stdout.fileno())
+        async_cb = self._on_async.__get__(self, UvStream)
+        self._async = pyuv.Async(self._loop, async_cb)
+        self._interrupted = False
 
     """
     Called when the libuv stream is connected
@@ -42,6 +46,7 @@ class UvStream(object):
             return
         self._connected = True
         self._read_stream = self._write_stream = stream
+        self._read_stream.start_read(self._read_cb)
 
     """
     Called when data is read from the libuv stream
@@ -50,10 +55,19 @@ class UvStream(object):
         if error:
             self._error = IOError(pyuv.errno.strerror(error))
             return
-        if not data:
+        elif not data:
             self._error = IOError('EOF')
             return
-        self._data = data
+        else:
+            self._data.append(data)
+        self._loop.stop()
+
+    """
+    Called when the async handle is fired
+    """
+    def _on_async(self, handle):
+        self._interrupted = True
+        self._loop.stop()
 
     """
     Called when a timeout occurs
@@ -70,6 +84,7 @@ class UvStream(object):
             self._error = IOError(pyuv.errno.strerror(error))
             return
         self._written = True
+        self._loop.stop()
     
     """
     Runs the event loop until a certain condition
@@ -102,15 +117,14 @@ class UvStream(object):
     def read(self, timeout=None):
         # first ensure the stream is connected
         self._run(lambda: self._connected)
-        # start reading
-        self._read_stream.start_read(self._read_cb)
         # wait until some data is read
-        self._run(lambda: self._data, timeout)
-        # stop reading
-        self._read_stream.stop_read()
+        self._run(lambda: self._interrupted or len(self._data), timeout)
+        if self._interrupted:
+            # Always reset this
+            self._interrupted = False
+            return
         # return a chunk of data
-        rv = self._data
-        self._data = None
+        rv = self._data.popleft()
         return rv
 
     """
@@ -126,3 +140,8 @@ class UvStream(object):
         # wait for the flag
         self._run(lambda: self._written)
 
+    """
+    Interrupts a `read` call from another thread.
+    """
+    def interrupt(self):
+        self._async.send()
